// return the difference of two isosurface tables

// Copyright (c) 2004 by Rephael Wenger, The Ohio State University 

// Author: Rephael Wenger

#include <fstream>
#include <iostream>
#include <vector>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "isotable.h"

using namespace std;
using namespace ISOTABLE;

// global variables
char * fname[2];

// routines
void diff_table_polyhedra(const ISOSURFACE_TABLE isotable[2]);
void parse_command_line(int argc, char **argv);
void convert(const ISOSURFACE_TABLE & isotable, 
	     const int it, const int is, bool * ebits);
bool equals(const bool * ebits0, const bool * ebits1, const int nume);

void usage_error();
void help_msg();

// create version of ISOSURFACE_TABLE with no constructor parameters
class ISOSURFACE_TABLE2:public ISOSURFACE_TABLE {
  public:
    ISOSURFACE_TABLE2():ISOSURFACE_TABLE(3) {};
};

int main(int argc, char **argv)
{
  parse_command_line(argc, argv);

  try {

    ISOSURFACE_TABLE2 isotable[2];
    ifstream in0(fname[0]);
    ifstream in1(fname[1]);

    isotable[0].ReadText(in0);
    isotable[1].ReadText(in1);

    for (int i = 0; i < 2; i++) {
      if (!isotable[i].CheckTable()) {
	cerr << "Error found in isosurface table "
	     << fname[i] << "." << endl;
	exit(12);
      };
    };

    if (isotable[0].Dimension() != isotable[1].Dimension()) {
      cerr << "Table dimension mismatch:" << endl;
      for (int i = 0; i < 2; i++) 
	cerr << "  " << fname[i] << " has dimension " 
	     << isotable[i].Dimension() << "." << endl;
      exit(20);
    };

    diff_table_polyhedra(isotable);

    if (isotable[0].NumTableEntries() != 
	isotable[1].NumTableEntries()) {
      cerr << "Table size mismatch:" << endl;
      for (int i = 0; i < 2; i++) 
	cerr << "  " << fname[i] << " has " 
	     << isotable[i].NumTableEntries() << " entries." << endl;
      exit(20);
    };

    bool * ebits[2];
    for (int i = 0; i < 2; i++) {
      ebits[i] = new bool[isotable[i].Polyhedron().NumEdges()];
    };

    for (int it = 0; 
	 it < isotable[0].NumTableEntries(); it++) {
      if (isotable[0].NumSimplices(it) != isotable[1].NumSimplices(it)) {
	cerr << "Table entry " << it << " mismatch:" << endl;
	for (int i = 0; i < 2; i++) 
	  cerr << "  " << fname[i] << " has " 
	       << isotable[i].NumSimplices(it) << " simplices." << endl;
	continue;
      };

      for (int is0 = 0; is0 < isotable[0].NumSimplices(it); is0++) {
	bool found_matching_simplex = false;
	convert(isotable[0], it, is0, ebits[0]);
	for (int is1 = 0; is1 < isotable[1].NumSimplices(it); is1++) {
	  convert(isotable[1], it, is1, ebits[1]);
	  if (equals(ebits[0], ebits[1], 
		     isotable[0].Polyhedron().NumEdges())) {
	    found_matching_simplex = true;
	  };
	};
	if (!found_matching_simplex) {
	  cerr << "Table entry " << it << " mismatch:" << endl;
	  cerr << "  " << fname[0] << ", simplex " << is0
	       << " has no match in " << fname[1] << "." << endl;
	  continue;
	};
      };
    };

    for (int i = 0; i < 2; i++)
      delete [] ebits[i];
  }
  catch (ISOSURFACE_TABLE_ERROR error) {
    cerr << "Programming error." << endl;
    cerr << error.Msg() << endl;
    cerr << "Exiting." << endl;
    exit(20);
  };
}

void diff_table_polyhedra(const ISOSURFACE_TABLE isotable[2])
// output differences in isosurface table polyhedra
// halt if vertex/edge differences are found
// do not halt if facet differences are found
{
  int dimension = isotable[0].Polyhedron().Dimension();

  assert(isotable[0].Polyhedron().Dimension() ==
	 isotable[1].Polyhedron().Dimension());

  if (isotable[0].Polyhedron().NumVertices() != 
      isotable[1].Polyhedron().NumVertices()) {
    cerr << "Polyhedron number of vertices mismatch:" << endl;
    for (int i = 0; i < 2; i++) 
      cerr << "  " << fname[i] << " has " 
	   << isotable[i].Polyhedron().NumVertices() 
	   << "vertices." << endl;
    exit(20);
  };

  if (isotable[0].Polyhedron().NumEdges() != 
      isotable[1].Polyhedron().NumEdges()) {
    cerr << "Polyhedron number of edges mismatch:" << endl;
    for (int i = 0; i < 2; i++) 
      cerr << "  " << fname[i] << " has " 
	   << isotable[i].Polyhedron().NumEdges() << " edges." << endl;
    exit(20);
  };

  for (int iv = 0; iv < isotable[0].Polyhedron().NumVertices(); iv++) {
    for (int ic = 0; ic < dimension; ic++) {
      int coord[2];
      coord[0] = isotable[0].Polyhedron().VertexCoord(iv, ic);
      coord[1] = isotable[1].Polyhedron().VertexCoord(iv, ic);
      if (coord[0] != coord[1]) {
	cerr << "Polyhedron vertex coordinate mismatch:" << endl;
	for (int i = 0; i < 2; i++) 
	  cerr << "  " << fname[i] << ", vertex " << iv
	       << ", coordinate " << ic << " = " 
	       << coord[i] << "." << endl;
	exit(20);
      };
    };
  };

  for (int ie = 0; ie < isotable[0].Polyhedron().NumEdges(); ie++) {
    int iend[2][2];

    for (int i = 0; i < 2; i++)
      for (int j = 0; j < 2; j++) {
	iend[i][j] = isotable[i].Polyhedron().EdgeEndpoint(ie, j);
      };

    if ((iend[0][0] != iend[1][0] && iend[0][0] != iend[1][1]) ||
	(iend[0][1] != iend[1][0] && iend[0][1] != iend[1][1])) {
      cerr << "Polyhedron edge mismatch:" << endl;
      for (int i = 0; i < 2; i++) 
	cerr << "  " << fname[i] << ", edge " << ie
	     << " = (" << iend[i][0] << ", " << iend[i][1] << ")." << endl;
      exit(20);
    };
  };

  // output facet differences
  // do not halt if facet differences are found
  if (isotable[0].Polyhedron().NumFacets() != 
      isotable[1].Polyhedron().NumFacets()) {
    cerr << "Polyhedron number of facets mismatch:" << endl;
    for (int i = 0; i < 2; i++) 
      cerr << "  " << fname[i] << " has " 
	   << isotable[i].Polyhedron().NumFacets() << " facets." << endl;
    return;
  };

  for (int jf = 0; jf < isotable[0].Polyhedron().NumFacets(); jf++) {
    if (isotable[0].Polyhedron().NumFacetVertices(jf) != 
	isotable[1].Polyhedron().NumFacetVertices(jf)) {
      cerr << "Polyhedron number of facet vertices mismatch:" << endl;
      for (int i = 0; i < 2; i++) 
	cerr << "  " << fname[i] << ", facet " << jf
	     << " has " << isotable[i].Polyhedron().NumFacetVertices(jf)
	     << " vertices." << endl;
      continue;
    };

    if (isotable[0].Polyhedron().Facet(jf) != 
	isotable[1].Polyhedron().Facet(jf)) {
      cerr << "Polyhedron facet vertices mismatch:" << endl;
      for (int iv = 0; iv < isotable[0].Polyhedron().NumVertices(); 
	   iv++) {
	if (isotable[0].Polyhedron().FacetVertexFlag(jf, iv) !=
	    isotable[1].Polyhedron().FacetVertexFlag(jf, iv)) {
	  for (int i = 0; i < 2; i++) 
	    if (isotable[i].Polyhedron().FacetVertexFlag(jf, iv)) 
	      cerr << "  " << fname[i] << ", facet " << jf
		   << " has vertex " << iv << "." << endl;
	    else
	      cerr << "  " << fname[i] << ", facet " << jf
		   << " does not have vertex " << iv << "." << endl;
	};
      };
    };
  };

}

void convert(const ISOSURFACE_TABLE & isotable, 
	     const int it, const int is, bool * ebits)
{
  for (int i = 0; i < isotable.Polyhedron().NumEdges(); i++) {
    ebits[i] = false;
  };

  for (int iv = 0; iv < isotable.NumVerticesPerSimplex(); iv++) {
    int ie = isotable.SimplexVertex(it, is, iv);
    ebits[ie] = true;
  };
}

bool equals(const bool * ebits0, const bool * ebits1, const int nume)
{
  for (int ie = 0; ie < nume; ie++) {
    if (ebits0[ie] != ebits1[ie]) {
      return(false);
    };
  };
  return(true);
}

void parse_command_line(int argc, char **argv)
{
  int iarg = 1;
  while (iarg < argc) {
    if (strcmp(argv[1], "-h") == 0) 
      { help_msg(); }
    else
      { break; }
  };

  if (argc != 3) { usage_error(); };

  fname[0] = argv[1];
  fname[1] = argv[2];
}

void usage_error()
{
  cerr << "Usage: diffisotable [-h] {file1} {file2}" << endl;
  exit(10);
}

void help_msg()
{
  cout << "diffisotable - output the difference of two isosurface tables" << endl;
  cout << "  Output differences in polyhedra and in table entries." << endl;
  cout << "  Ignore simplex order in table entries." << endl;
  cout << "  Ignore vertex order in comparing isosurface simplices." << endl;
  cout << endl;
  cout << "Usage: diffisotable [-h] {file1} {file2}" << endl;
  cout << "  -h : Print this help message (and exit.)" << endl;
  exit(0);
}
